From cb7b59474a2ca6b64474330ec2bb279c092f661a Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Petr=20=C5=A0pa=C4=8Dek?= <petr.spacek@nic.cz>
Date: Fri, 12 Apr 2019 13:34:49 +0200
Subject: [PATCH] http/h2_stream: ignore delayed RST_STREAM frames

Fixes: #145
---
 http/h2_stream.lua      |  3 +++
 spec/h2_stream_spec.lua | 37 +++++++++++++++++++++++++++++++++++++
 2 files changed, 40 insertions(+)

diff --git a/http/h2_stream.lua b/http/h2_stream.lua
index abd3899..4d34139 100644
--- a/http/h2_stream.lua
+++ b/http/h2_stream.lua
@@ -690,6 +690,9 @@ frame_handlers[frame_types.RST_STREAM] = function(stream, flags, payload, deadli
 	end
 	if stream.state == "idle" then
 		return nil, h2_errors.PROTOCOL_ERROR:new_traceback("'RST_STREAM' frames MUST NOT be sent for a stream in the 'idle' state"), ce.EILSEQ
+	elseif stream.state == "closed" then
+		-- probably a delayed RST_STREAM, ignore
+		return true
 	end
 
 	local err_code = sunpack(">I4", payload)
diff --git a/spec/h2_stream_spec.lua b/spec/h2_stream_spec.lua
index 3213cd0..1cfed79 100644
--- a/spec/h2_stream_spec.lua
+++ b/spec/h2_stream_spec.lua
@@ -124,6 +124,43 @@ describe("http.h2_stream", function()
 			assert_loop(cq, TEST_TIMEOUT)
 			assert.truthy(cq:empty())
 		end)
+		it("ignores delayed RST_STREAM on already closed stream", function()
+			local s, c = new_pair()
+			local cq = cqueues.new()
+			cq:wrap(function()
+				local client_stream = c:new_stream()
+				local req_headers = new_headers()
+				req_headers:append(":method", "GET")
+				req_headers:append(":scheme", "http")
+				req_headers:append(":path", "/")
+				req_headers:append(":authority", "example.com")
+				assert(client_stream:write_headers(req_headers, true))
+				assert(client_stream:get_headers())
+				assert("closed", client_stream.state)
+				-- both sides now have stream in closed state
+				-- send server a RST_STREAM: it should get ignored
+				assert(client_stream:rst_stream("post-closed rst_stream"))
+				assert(c:close())
+			end)
+			cq:wrap(function()
+				local stream = assert(s:get_next_incoming_stream())
+				assert(stream:get_headers())
+				local res_headers = new_headers()
+				res_headers:append(":status", "200")
+				assert(stream:write_headers(res_headers, true))
+				-- both sides now have stream in closed state
+				assert("closed", stream.state)
+				-- process incoming frames until EOF (i.e. drain RST_STREAM)
+				-- the RST_STREAM frame should be ignored.
+				assert(s:loop())
+				assert(s:close())
+			end)
+			cq:wrap(function()
+				assert(s:loop())
+			end)
+			assert_loop(cq, TEST_TIMEOUT)
+			assert.truthy(cq:empty())
+		end)
 	end)
 	describe("push_promise", function()
 		it("permits a simple push promise from server => client", function()
